/* gAlan - Graphical Audio Language
 * Copyright (C) 1999 Tony Garnock-Jones
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <gdk/gdk.h>
#include <gtk/gtk.h>

#include "global.h"
#include "generator.h"
#include "comp.h"
#include "control.h"
#include "gencomp.h"

#define GENERATOR_CLASS_NAME	"voice"
#define GENERATOR_CLASS_PATH	"Sources/Sampled Voice"

typedef struct Data {
  SAMPLE *sample;
  int len;
} Data;

PRIVATE int init_instance(Generator *g) {
  Data *data = malloc(sizeof(Data));

  g_return_val_if_fail(data != NULL, 0);
  g->data = data;

  data->sample = NULL;
  data->len = 0;

  return 1;
}

PRIVATE void destroy_instance(Generator *g) {
  Data *data = g->data;

  if (data->sample)
    free(data->sample);

  free(g->data);
}

PRIVATE void unpickle_instance(Generator *g, ObjectStoreItem *item, ObjectStore *db) {
  Data *data = malloc(sizeof(Data));
  int i, len;
  gint16 *buf;

  g_return_if_fail(data != NULL);
  g->data = data;

  data->len = len = objectstore_item_get_integer(item, "voice_length", 0);
  objectstore_item_get_binary(item, "voice_sample", (void **) &buf);

  data->sample = malloc(sizeof(SAMPLE) * len);
  g_return_if_fail(data->sample != NULL);
  for (i = 0; i < len; i++)
    data->sample[i] = ((gint16) g_ntohs(buf[i])) / 32768.0;
}

PRIVATE void pickle_instance(Generator *g, ObjectStoreItem *item, ObjectStore *db) {
  Data *data = g->data;
  int bytelength = sizeof(gint16) * data->len;
  gint16 *buf = malloc(bytelength);
  int i;

  g_return_if_fail(buf != NULL);

  objectstore_item_set_integer(item, "voice_length", data->len);
  for (i = 0; i < data->len; i++)
    buf[i] = g_htons(MAX(-32768, MIN(32767, 32768 * data->sample[i])));
  objectstore_item_set(item, "voice_sample",
		       objectstore_datum_new_binary(bytelength, (void *) buf));
  free(buf);
}

PRIVATE SAMPLETIME output_range(Generator *g, OutputSignalDescriptor *sig) {
  return ((Data *) g->data)->len;
}

PRIVATE gboolean output_generator(Generator *g, OutputSignalDescriptor *sig,
				  SAMPLETIME offset, SAMPLE *buf, int buflen) {
  Data *data = g->data;
  int len, sil;

  if (data->len == 0 || offset >= data->len)
    return FALSE;

  len = MIN(MAX(data->len - offset, 0), buflen);
  if (len > 0)
    memcpy(buf, &data->sample[offset], len * sizeof(SAMPLE));

  sil = buflen - len;
  memset(&buf[len], 0, sil * sizeof(SAMPLE));
  return TRUE;
}

PRIVATE void load_new_sample(GtkWidget *widget, GtkWidget *fs) {
  Generator *g = gtk_object_get_data(GTK_OBJECT(fs), "Generator");
  Data *data = g->data;
  FILE *f;
  long pos;

  f = fopen(gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)), "rb");

  if (f == NULL)
    return;

  if (data->sample != NULL)
    free(data->sample);

  fseek(f, 0, SEEK_END);
  pos = ftell(f);
  fseek(f, 0, SEEK_SET);

  data->len = pos / sizeof(gint16);
  data->sample = malloc(sizeof(SAMPLE) * data->len);

  if (data->sample == NULL) {
    fclose(f);
    return;
  }

  for (pos = 0; pos < data->len; pos++) {
    gint16 i;

    fread(&i, sizeof(i), 1, f);
    data->sample[pos] = i / 32768.0;
  }

  fclose(f);
  gtk_widget_destroy(fs);	/* %%% should this be gtk_widget_hide? uber-paranoia */
}

PRIVATE void propgen(Component *c, Generator *g) {
  GtkWidget *fs = gtk_file_selection_new("Sample File");

  gtk_object_set_data(GTK_OBJECT(fs), "Generator", g);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fs)->ok_button), "clicked",
		     GTK_SIGNAL_FUNC(load_new_sample), fs);
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(fs)->cancel_button), "clicked",
			    GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(fs));

  gtk_widget_show(fs);
}

PRIVATE OutputSignalDescriptor output_sigs[] = {
  { "Output", SIG_FLAG_RANDOMACCESS, { NULL, { output_range, output_generator } } },
  { NULL, }
};

PRIVATE void setup_class(void) {
  GeneratorClass *k = gen_new_generatorclass(GENERATOR_CLASS_NAME, 0, 0,
					     NULL, output_sigs, NULL,
					     init_instance, destroy_instance,
					     unpickle_instance, pickle_instance);

  gencomp_register_generatorclass(k, GENERATOR_CLASS_PATH, NULL, propgen);
}

PUBLIC void init_plugin_voice(void) {
  setup_class();
}
